

/*
**
**  $VER: dtrequester.c 1.2 (27.3.97)
**  datatypes.library/Examples/dtrequester
**
**  Small datatypes browser;
**  How to embed a DataType object within an Intuition
**  requester or window.
**
**  Written 1996/97 by Roland 'Gizzy' Mainz
**
*/

/* amiga includes */
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/libraries.h>
#include <dos/dos.h>
#include <dos/rdargs.h>
#include <dos/stdio.h>
#include <graphics/gfx.h>
#include <graphics/displayinfo.h>
#include <intuition/intuition.h>
#include <intuition/icclass.h>
#include <intuition/imageclass.h>
#include <datatypes/datatypes.h>
#include <datatypes/datatypesclass.h>
#include <datatypes/textclass.h>
#include <datatypes/pictureclass.h>
#include <datatypes/soundclass.h>
#include <datatypes/animationclass.h>
#include <workbench/icon.h>

/* amiga prototypes */
#include <clib/macros.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
#include <clib/exec_protos.h>
#include <clib/utility_protos.h>
#include <clib/dos_protos.h>
#include <clib/graphics_protos.h>
#include <clib/intuition_protos.h>
#include <clib/datatypes_protos.h>
#include <clib/icon_protos.h>

/* amiga pragmas */
#include <pragmas/exec_pragmas.h>
#include <pragmas/utility_pragmas.h>
#include <pragmas/dos_pragmas.h>
#include <pragmas/graphics_pragmas.h>
#include <pragmas/intuition_pragmas.h>
#include <pragmas/datatypes_pragmas.h>
#include <pragmas/icon_pragmas.h>
#include <pragmas/alib_pragmas.h> /* tagcall stubs */

/* ansi includes */
#include <string.h>
#include <limits.h>

/* macros */
#define IsDTMethodSupported( o, id ) ((BOOL)FindMethod(GetDTMethods( (o) ), (id) ))


/*****************************************************************************/

extern struct Library *SysBase,
                      *DOSBase;

/*****************************************************************************/

struct Library *UtilityBase;
struct Library *GfxBase;
struct Library *IntuitionBase;
struct Library *DataTypesBase;
struct Library *IconBase;

/*****************************************************************************/

struct FrameInfo gfri = { 0 };

/*****************************************************************************/

/* local prototypes */
static void  PrintErrorMsg( ULONG, STRPTR );

static void  PrintDTTriggerMethods( Object * );
static ULONG DTGoto( Object *, struct Window *, struct Requester *, STRPTR, struct TagItem * );
static ULONG DoDTFrameBoxSpecify( Object *, struct Window *, struct Requester *, struct FrameInfo * );
static void  PrintFrameInfo( struct FrameInfo * );
static BOOL  FillTriggerData( struct dtTrigger *, UBYTE *, ULONG );
static void  FlushInOut( void );


/*****************************************************************************/

#define TEMPLATE "NAME/A,REQUESTER/S,PROGRESSIVE/S,DATATYPE/K,PUBSCREEN/K"
#define OPT_NAME        (0)
#define OPT_REQUESTER   (1)
#define OPT_PROGRESSIVE (2)
#define OPT_DATATYPE    (3)
#define OPT_PUBSCREEN   (4)
#define NUM_OPTS        (5)

/*****************************************************************************/

#define IDCMP_FLAGS (IDCMP_CLOSEWINDOW | IDCMP_VANILLAKEY | IDCMP_RAWKEY | IDCMP_GADGETDOWN | IDCMP_GADGETUP | IDCMP_IDCMPUPDATE)

/*****************************************************************************/

static
void PrintErrorMsg( ULONG errnum, STRPTR name )
{
    STRPTR format;
    TEXT   errbuff[ 120 ];

    if( errnum >= DTERROR_UNKNOWN_DATATYPE )
    {
      format = GetDTString( errnum );

      if( (ULONG)(strlen( format ) + strlen( name )) < 110UL )
      {
        sprintf( errbuff, format, name );
      }
      else
      {
        /* buffer overflow */
        errbuff[ 0 ] = 0U;
      }
    }
    else
    {
      Fault( errnum, NULL, errbuff, sizeof( errbuff ) );
    }

    Printf( "%s\nerror #%ld\n", errbuff, errnum );
}

/*****************************************************************************/

void main( void )
{
    Object               *dto;

    ULONG                 modeid    = (ULONG)INVALID_ID;
    BOOL                  useScreen = FALSE;
    struct Screen        *scr;
    struct Window        *win;
    BOOL                  going     = TRUE;

    struct IntuiMessage  *imsg;
    ULONG                 sigmask,
                          signals;

    struct TagItem       *tstate,
                         *ti;
    ULONG                 tidata;
    ULONG                 errnum;

    LONG                  options[ NUM_OPTS ];
    struct RDArgs        *rdargs;

    TEXT                  stm_data[ 256 ];

    /* dtTrigger is used by IDCMP_VANILLAKEY/IDCMP_RAWKEY */
    struct dtTrigger dtt =
    {
      DTM_TRIGGER,
      NULL,
      0UL,
      NULL
    };

    /* Parse the arguments.  Note that this simple example assumes
     * that it was started from the shell.
     */
    memset( options, 0, sizeof( options ) );

    if( rdargs = ReadArgs( TEMPLATE, options, NULL ) )
    {
      /* Open the other libraries */
      UtilityBase   = OpenLibrary( "utility.library",   39UL );
      GfxBase       = OpenLibrary( "graphics.library",  39UL );
      IntuitionBase = OpenLibrary( "intuition.library", 39UL );
      DataTypesBase = OpenLibrary( "datatypes.library", 45UL );
      IconBase      = OpenLibrary( "icon.library",      39UL );

      if( UtilityBase && GfxBase && IntuitionBase && DataTypesBase && IconBase )
      {
        struct DataType *dtn = NULL;

        if( options[ OPT_DATATYPE ] )
        {
          if( !(dtn = ObtainDataTypeA( DTST_RAM, (APTR)options[ OPT_DATATYPE ], NULL )) )
          {
            Printf( "Can't find given datatype '%s'\n", (APTR)options[ OPT_DATATYPE ] );
          }
        }

        /* Get a DataType object */
        if( dto = NewDTObject( (APTR)options[ OPT_NAME ], GA_RelVerify,    TRUE,
                                                          GA_Immediate,    TRUE,
                                                          DTA_DataType,    dtn,
                                                          DTA_Progressive, (ULONG)options[ OPT_PROGRESSIVE ],
                                                          TAG_DONE ) )
        {
          struct dtFrameBox     dtf = { 0 };
          struct FrameInfo      fri = { 0 };
          STRPTR                name;

          PrintDTTriggerMethods( dto );

          /* Get information about the object */
          if( GetDTAttrs( dto,
                          /* Get the name of the object */
                          DTA_ObjName,      (ULONG)(&name),

                          /* Get the mode ID for graphical objects */
                          PDTA_ModeID,      (ULONG)(&modeid),
                          TAG_DONE ) )
          {
            /* Display any information we obtained */
            if( name )
            {
              Printf( "opened \"%s\"\n", name );
            }

            /* Show the mode ID */
            Printf( "mode ID %08lx\n", modeid );
          }

          Printf( "Asking the object what kind of environment it needs...\n" );

          /* Ask the object what kind of environment it needs */
          dtf . MethodID          = DTM_FRAMEBOX;
          dtf . dtf_FrameInfo     = (&fri);
          dtf . dtf_ContentsInfo  = (&fri);
          dtf . dtf_SizeFrameInfo = sizeof( struct FrameInfo );

          if( DoDTMethodA( dto, NULL, NULL, (Msg)(&dtf) ) )
          {
            /* Really success ? */
            if( fri . fri_Dimensions . Depth )
            {
              Printf( "DTM_FRAMEBOX returned\n" );

              PrintFrameInfo( (&fri) );

              if( (fri . fri_PropertyFlags) & DIPF_IS_HAM )
              {
                useScreen = TRUE;
              }

              if( (fri . fri_PropertyFlags) & DIPF_IS_EXTRAHALFBRITE )
              {
                useScreen = TRUE;
              }

              if( (fri . fri_PropertyFlags == 0UL) && (modeid & HAM_KEY) && (modeid != INVALID_ID) )
              {
                useScreen = TRUE;
              }
            }
            else
            {
              Printf( "DTM_FRAMEBOX failed\n" );
            }
          }
          else
          {
            Printf( "DTM_FRAMEBOX failed / couldn't obtain environment information\n" );
          }

          if( useScreen )
          {
            Printf( "this object requires a private screen\n" );
          }

          /* Get a lock on the specified public screen */
          if( scr = LockPubScreen( (STRPTR)options[ OPT_PUBSCREEN ] ) )
          {
            Object *reqframe;

            if( reqframe = (Object *)NewObject( NULL, FRAMEICLASS, IA_FrameType, FRAME_DEFAULT, TAG_DONE ) )
            {
              struct Requester  req       = { 0 };
              struct Requester *requester = NULL;
              struct IBox       domain;
              ULONG             minwidth  = 0UL,
                                minheight = 0UL,
                                nomwidth  = 0UL,
                                nomheight = 0UL,
                                maxwidth  = SHRT_MAX,
                                maxheight = SHRT_MAX;

              if( options[ OPT_REQUESTER ] )
              {
                requester = &req;

                /* Init struct Requester (autodocs says that this is useless, but ...) */
                InitRequester( (&req) );

                req . Flags |= (USEREQIMAGE | NOREQBACKFILL | NOISYREQ);
                req . ReqImage = (struct Image *)reqframe;
              }

              /* Obtain minimum object box */
              if( DoDTDomainA( dto, NULL, NULL, NULL, GDOMAIN_MINIMUM, (&domain), NULL ) )
              {
                minwidth  = domain . Width;
                minheight = domain . Height;

                Printf( "min box %d %d %d %d\n", domain );
              }

              /* Obtain nominal object box */
              if( DoDTDomainA( dto, NULL, NULL, NULL, GDOMAIN_NOMINAL, (&domain), NULL ) )
              {
                nomwidth  = domain . Width;
                nomheight = domain . Height;

                Printf( "nom box %d %d %d %d\n", domain );
              }

              /* Obtain maximum object box */
              if( DoDTDomainA( dto, NULL, NULL, NULL, GDOMAIN_MAXIMUM, (&domain), NULL ) )
              {
                maxwidth  = domain . Width;
                maxheight = domain . Height;

                Printf( "max box %d %d %d %d\n", domain );
              }

              /* Make sure we have the right dimensions */
              minwidth  = MAX( minwidth,  80 );
              minheight = MAX( minheight, 40 );
              nomwidth  = ((nomwidth)?(nomwidth):(600UL));
              nomheight = ((nomheight)?(nomheight):(175UL));

              /* Open the window */
              if( win = OpenWindowTags( NULL, WA_InnerWidth,    (nomwidth  + ((requester)?(80):(4))),
                                              WA_InnerHeight,   (nomheight + ((requester)?(40):(2))),
                                              WA_Title,         name,
                                              WA_IDCMP,         IDCMP_FLAGS,
                                              WA_DragBar,       TRUE,
                                              WA_DepthGadget,   TRUE,
                                              WA_CloseGadget,   (!requester),
                                              WA_SizeGadget,    (!requester),
                                              WA_AutoAdjust,    TRUE,
                                              WA_SimpleRefresh, TRUE,
                                              WA_BusyPointer,   TRUE,
                                              WA_Activate,      TRUE,
                                              WA_RptQueue,      50UL,
                                              WA_PubScreen,     scr,
                                              TAG_DONE ) )
              {
                /* Set window limits, based on min/max dimensions of object */
                WindowLimits( win,
                              ((win -> BorderLeft) + (win -> BorderRight)  + 4L + minwidth),
                              ((win -> BorderTop)  + (win -> BorderBottom) + 2L + minheight),
                              ((win -> BorderLeft) + (win -> BorderRight)  + 4L + maxwidth),
                              ((win -> BorderTop)  + (win -> BorderBottom) + 2L + maxheight) );

                if( requester )
                {
                  /* layout requester */
                  req . LeftEdge   = win -> BorderLeft + 20UL;
                  req . TopEdge    = win -> BorderTop  + 10UL;
                  req . Width      = (win -> Width)  - ((win -> BorderLeft) + (win -> BorderRight)  + 40UL);
                  req . Height     = (win -> Height) - ((win -> BorderTop)  + (win -> BorderBottom) + 20UL);

                  SetAttrs( reqframe,
                            IA_Left,   0UL,
                            IA_Top,    0UL,
                            IA_Width,  (req . Width),
                            IA_Height, (req . Height),
                            TAG_DONE );
                }

                if( (requester)?(Request( requester, win )):(TRUE) )
                {
                  BOOL inwindow;

                  DoDTFrameBoxSpecify( dto, win, requester, (&gfri) );

                  if( requester )
                  {
                    /* Set the dimensions of the DataType object in the requester */
                    SetDTAttrs( dto, NULL, NULL, GA_Left,      20UL,
                                                 GA_Top,       10UL,
                                                 GA_Width,     ((req . Width)  - 40UL),
                                                 GA_Height,    ((req . Height) - 20UL),
                                                 ICA_TARGET,   ICTARGET_IDCMP,
                                                 TAG_DONE );
                  }
                  else
                  {
                    /* Set the dimensions of the DataType object in a window */
                    SetDTAttrs( dto, NULL, NULL, GA_Left,      ((win -> BorderLeft) + 2UL),
                                                 GA_Top,       ((win -> BorderTop)  + 1UL),
                                                 GA_RelWidth,  (-((LONG)((win -> BorderLeft) + (win -> BorderRight)  + 4UL))),
                                                 GA_RelHeight, (-((LONG)((win -> BorderTop)  + (win -> BorderBottom) + 2UL))),
                                                 ICA_TARGET,   ICTARGET_IDCMP,
                                                 TAG_DONE );
                  }

                  /* Add the object to the requester */
                  AddDTObject( win, requester, dto, (-1L) );

                  /* Refresh the DataType object */
                  RefreshDTObjectA( dto, win, requester, NULL );

                  /* Object is in window */
                  inwindow = TRUE;

                  /* Wait for window events or user kill */
                  sigmask = ((1UL << win -> UserPort -> mp_SigBit) | SIGBREAKF_CTRL_C);

                  /* Keep going until we're told to stop */
                  while( going )
                  {
                    /* Wait for an event */
                    signals = Wait( sigmask );

                    /* Did we get a break signal */
                    if( signals & SIGBREAKF_CTRL_C )
                    {
                      going = FALSE;
                    }

                    /* Pull Intuition messages */
                    while( imsg = (struct IntuiMessage *)GetMsg( (win -> UserPort) ) )
                    {
                      /* Handle each message */
                      switch( imsg -> Class )
                      {
                        case IDCMP_CLOSEWINDOW:
                        {
                            going = FALSE;
                        }
                            break;

                        case IDCMP_VANILLAKEY:
                        {
                            switch( imsg -> Code )
                            {
/* Quit application */
                              case 'Q':
                              case 'q':
                              case  27: /* ESC */
                              {
                                  Printf( "Quit application\n" );

                                  going = FALSE;
                              }
                                  break;

/* Add object into window/requester */
                              case '+':
                              {
                                  Printf( "Add object into window/requester\n" );

                                  if( inwindow == FALSE )
                                  {
                                    /* Add the object to the window */
                                    AddDTObject( win, requester, dto, (-1L) );

                                    /* Refresh the DataType object */
                                    RefreshDTObjectA( dto, win, requester, NULL );

                                    inwindow = TRUE;
                                  }
                                  else
                                  {
                                    Printf( "Object aleady in window/requester\n" );
                                  }
                              }
                                  break;

/* Remove object from window/requester */
                              case '-':
                              {
                                  Printf( "Remove object from window/requester\n" );

                                  if( inwindow )
                                  {
                                    /* Remove the object from the window */
                                    RemoveDTObject( win, dto );
                                    inwindow = FALSE;

                                    if( requester )
                                    {
                                      EraseRect( (requester -> ReqLayer -> rp),
                                                 20UL,
                                                 10UL,
                                                 ((requester -> Width)  - 20UL),
                                                 ((requester -> Height) - 10UL) );
                                    }
                                    else
                                    {
                                      EraseRect( (win -> RPort),
                                                 ((win -> BorderLeft) + 2UL),
                                                 ((win -> BorderTop)  + 1UL),
                                                 ((win -> Width)  - ((win -> BorderRight)  + 4UL)),
                                                 ((win -> Height) - ((win -> BorderBottom) + 2UL)) );
                                    }
                                  }
                                  else
                                  {
                                    Printf( "Object not in window/requester\n" );
                                  }
                              }
                                  break;

/* Start marking */
                              case 'B':
                              case 'b':
                              {
                                  Printf( "Start marking\n" );

                                  if( inwindow )
                                  {
                                    if( !StartDragSelect( dto ) )
                                    {
                                      Printf( "Can't start marking\n" );
                                    }
                                  }
                              }
                                  break;

/* Force refresh */
                              case 'R':
                              case 'r':
                              {
                                  Printf( "Force refresh\n" );

                                  if( inwindow )
                                  {
                                    RefreshDTObjectA( dto, win, requester, NULL );
                                  }
                              }
                                  break;

/* Copy to clipboard */
                              case 'C':
                              case 'c':
                              {
                                  Printf( "Copy to clipboard\n" );

                                  if( inwindow )
                                  {
                                    if( IsDTMethodSupported( dto, DTM_COPY ) )
                                    {
                                      DoDTMethod( dto, win, requester, DTM_COPY, NULL );
                                    }
                                    else
                                    {
                                      Printf( "DTM_COPY not supported\n" );
                                    }
                                  }
                              }
                                  break;

/* Clear selected area */
                              case ']':
                              {
                                  Printf( "Clear selected area\n" );

                                  if( inwindow )
                                  {
                                    if( IsDTMethodSupported( dto, DTM_CLEARSELECTED ) )
                                    {
                                      struct DTSpecialInfo *dtsi = ((struct DTSpecialInfo *)(((struct ExtGadget *)dto) -> SpecialInfo));

                                      if( (dtsi -> si_Flags) & DTSIF_HIGHLIGHT )
                                      {
                                        /* Clear selected area */
                                        DoDTMethod( dto, win, requester, DTM_CLEARSELECTED, NULL );

                                        /* Refresh the DataType object */
                                        RefreshDTObjectA( dto, win, requester, NULL );
                                      }
                                      else
                                      {
                                        Printf( "nothing selected\n" );
                                      }
                                    }
                                    else
                                    {
                                      Printf( "DTM_CLEARSELECTED not supported\n" );
                                    }
                                  }
                              }
                                  break;

/* Goto node */
                              case 'G':
                              case 'g':
                              {
                                  Printf( "Goto node\n" );

                                  if( inwindow )
                                  {
                                    if( IsDTMethodSupported( dto, DTM_GOTO ) )
                                    {
                                      TEXT nodebuffer[ 256 ];

                                      Printf( "enter node to go:" ); FlushInOut();

                                      if( ReadItem( nodebuffer, (sizeof( nodebuffer ) - 1), NULL ) )
                                      {
                                        DTGoto( dto, win, requester, nodebuffer, NULL );
                                      }

                                      /* flush input (see ReadItem autodoc) */
                                      FlushInOut();

                                      ActivateWindow( win );
                                    }
                                    else
                                    {
                                      Printf( "DTM_GOTO not supported\n" );
                                    }
                                  }
                              }
                                  break;

/* Invoke trigger method */
                              case 't':
                              case 'T':
                              {
                                  Printf( "Invoke trigger method\n" );

                                  if( inwindow )
                                  {
                                    if( IsDTMethodSupported( dto, DTM_TRIGGER ) )
                                    {
                                      struct DTMethod *dtm;

                                      if( dtm = GetDTTriggerMethods( dto ) )
                                      {
                                        TEXT namebuffer[ 256 ];

                                        Printf( "enter trigger method name:" ); FlushInOut();

                                        if( ReadItem( namebuffer, (sizeof( namebuffer ) - 1), NULL ) )
                                        {
                                          if( dtm = FindTriggerMethod( dtm, namebuffer, 0UL ) )
                                          {
                                            dtt . dtt_Function = dtm -> dtm_Method;

                                            if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                            {
                                              DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                            }
                                          }
                                          else
                                          {
                                            Printf( "trigger method not found\n" );
                                          }
                                        }
                                      }

                                      /* flush input (see ReadItem autodoc) */
                                      FlushInOut();

                                      ActivateWindow( win );
                                    }
                                    else
                                    {
                                      Printf( "DTM_TRIGGER not supported\n" );
                                    }
                                  }
                              }
                                  break;

/* Retrace */
                              case '/':
                              {
                                  Printf( "Retrace\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_RETRACE;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;

/* Next field */
                              case 9: /* TAB */
                              {
                                  Printf( "Next field\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_NEXT_FIELD;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;

/* Activate field */
                              case 13: /* ENTER */
                              {
                                  Printf( "Activate field\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_ACTIVATE_FIELD;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;

/* Previous page */
                              case '<':
                              {
                                  Printf( "Previous page\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_BROWSE_PREV;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;

/* Next page */
                              case '>':
                              {
                                  Printf( "Next page\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_BROWSE_NEXT;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;

/* Locate */
                              case 'L':
                              case 'l':
                              {
                                  Printf( "Locate\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_LOCATE;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;
                            }
                        }
                            break;

                        case IDCMP_RAWKEY:
                        {
                            switch( imsg -> Code )
                            {
/* Previous field */
                              case 66: /* shift-TAB */
                              {
                                  Printf( "Previous field\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_PREV_FIELD;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;

                              case 95: /* help */
                              {
/* Send command */
                                  Printf( "Send command\n" );

                                  if( inwindow )
                                  {
                                    dtt . dtt_Function = STM_COMMAND;

                                    if( FillTriggerData( (&dtt), stm_data, sizeof( stm_data )) )
                                    {
                                      DoDTMethodA( dto, win, requester, (Msg)(&dtt) );
                                    }
                                  }
                              }
                                  break;
                            }
                        }
                            break;

                        case IDCMP_IDCMPUPDATE:
                        {
                            struct TagItem *tags;

                            tags = tstate = imsg -> IAddress;

                            while( ti = NextTagItem( (&tstate) ) )
                            {
                              tidata = ti -> ti_Data;

                              switch( ti -> ti_Tag )
                              {
                                /* Change in busy state */
                                case DTA_Busy:
                                {
                                    Printf( "busy %lu\n", tidata );

                                    SetWindowPointer( win, WA_BusyPointer, tidata, TAG_DONE );
                                }
                                    break;

                                /* Error message */
                                case DTA_ErrorLevel:
                                {
                                    if( tidata )
                                    {
                                      errnum = GetTagData( DTA_ErrorNumber, 0UL, tags );
                                      PrintErrorMsg( errnum, (STRPTR)GetTagData( DTA_ErrorString, (ULONG)options[ OPT_NAME ], tags ) );
                                    }
                                }
                                    break;

                                /* Time to refresh (visuals and internal variables) */
                                case DTA_Sync:
                                {
                                    if( inwindow )
                                    {
                                      Printf( "DTA_Sync\n" );

                                      /* Refresh the DataType object */
                                      RefreshDTObjectA( dto, win, requester, NULL );

                                      /* Maybe we've new dimensions... */
                                      if( DoDTDomainA( dto, win, requester, NULL, GDOMAIN_MINIMUM, (&domain), NULL ) )
                                      {
                                        minwidth  = domain . Width;
                                        minheight = domain . Height;

                                        minwidth  = MAX( minwidth,  80 );
                                        minheight = MAX( minheight, 40 );

                                        Printf( "min box %d %d %d %d\n", domain );
                                      }

                                      /* dito. max. box */
                                      if( DoDTDomainA( dto, win, requester, NULL, GDOMAIN_MAXIMUM, (&domain), NULL ) )
                                      {
                                        maxwidth  = domain . Width;
                                        maxheight = domain . Height;

                                        Printf( "max box %d %d %d %d\n", domain );
                                      }

                                      /* Set new limits */
                                      WindowLimits( win,
                                                    ((win -> BorderLeft) + (win -> BorderRight)  + 4L + minwidth),
                                                    ((win -> BorderTop)  + (win -> BorderBottom) + 2L + minheight),
                                                    ((win -> BorderLeft) + (win -> BorderRight)  + 4L + maxwidth),
                                                    ((win -> BorderTop)  + (win -> BorderBottom) + 2L + maxheight) );
                                    }
                                }
                                    break;

                                case DTA_Title:
                                {
                                    SetWindowTitles( win, (STRPTR)tidata, (STRPTR)(~0UL) );
                                }
                                    break;
                              }
                            }
                        }
                            break;

                        case IDCMP_GADGETDOWN:
                        {
                            Printf( "IDCMP_GADGETDOWN %lx %lx\n", (imsg -> IAddress), (ULONG)(imsg -> Code) );
                        }
                            break;

                        case IDCMP_GADGETUP:
                        {
                            Printf( "IDCMP_GADGETUP %lx %lx\n", (imsg -> IAddress), (ULONG)(imsg -> Code) );
                        }
                            break;
                      }

                      /* Done with the message, so reply to it */
                      ReplyMsg( (&(imsg -> ExecMessage)) );
                    }
                  }

                  Printf( "Done...\n" );

                  if( inwindow )
                  {
                    /* Remove the object from the window */
                    RemoveDTObject( win, dto );
                  }

                  if( requester )
                  {
                    EndRequest( requester, win );
                  }
                }

                /* Close the window now */
                CloseWindow( win );
              }
              else
              {
                Printf( "Couldn't open window\n" );
              }

              DisposeObject( reqframe );
            }

            /* Unlock the screen */
            UnlockPubScreen( NULL, scr );
          }
          else
          {
            Printf( "Couldn't lock public screen\n" );
          }

          /* Dispose of the DataType object */
          DisposeDTObject( dto );
        }
        else
        {
          PrintErrorMsg( IoErr(), (STRPTR)options[ OPT_NAME ] );
        }

        /* Release datatype */
        ReleaseDataType( dtn );
      }
      else
      {
        Printf( "Couldn't open a shared library\n" );
      }

      /* Close the libraries */
      CloseLibrary( IconBase );
      CloseLibrary( DataTypesBase );
      CloseLibrary( IntuitionBase );
      CloseLibrary( GfxBase );
      CloseLibrary( UtilityBase );

      FreeArgs( rdargs );
    }
    else
    {
      /* ReadArgs failed */
      PrintErrorMsg( IoErr(), "DTRequester" );
    }
}


static
void PrintDTTriggerMethods( Object *dto )
{
    struct DTMethod *dtm;

    if( dto )
    {
      if( dtm = GetDTTriggerMethods( dto ) )
      {
        Printf( "Follwoing trigger methods are defined:\n" );

        while( (dtm -> dtm_Method) != STM_DONE )
        {
          Printf( "Label: '%10.10s' ",   (dtm -> dtm_Label) );
          Printf( "Command: '%10.10s' ", (dtm -> dtm_Command) );
          Printf( "Method: %10.10lu ",   (dtm -> dtm_Method) );

          switch( GetDTTriggerMethodDataFlags( (dtm -> dtm_Method)) )
          {
            case STMD_VOID:    Printf( "(VOID)"     ); break;
            case STMD_ULONG:   Printf( "(ULONG)"    ); break;
            case STMD_STRPTR:  Printf( "(STRPTR)"   ); break;
            case STMD_TAGLIST: Printf( "(TAGLIST)"  ); break;
            default:           Printf( "--UNKOWN--" ); break;
          }

          Printf( "\n" );

          dtm++;
        }
      }
      else
      {
        Printf( "No trigger methods defined\n" );
      }
    }
}


static
ULONG DTGoto( Object *o, struct Window *win, struct Requester *req, STRPTR node, struct TagItem *Attrs )
{
    struct dtGoto dtg;

    dtg . MethodID     = DTM_GOTO;
    dtg . dtg_GInfo    = NULL;
    dtg . dtg_NodeName = node;
    dtg . dtg_AttrList = Attrs;

    return( DoDTMethodA( o, win, req, (Msg)(&dtg) ) );
}


/* Tell the object about the environment it will be going into */
static
ULONG DoDTFrameBoxSpecify( Object *dto, struct Window *win, struct Requester *req, struct FrameInfo *fri )
{
    if( dto && fri )
    {
      struct Screen       *scr;
      struct DisplayInfo   di;
      ULONG                modeid;
      struct dtFrameBox    dtf;

      scr = win -> WScreen;

      /* Get the display information */
      modeid = GetVPModeID( (&(scr -> ViewPort)) );
      (void)GetDisplayInfoData( NULL, (APTR)(&di), sizeof( struct DisplayInfo ), DTAG_DISP, modeid );

      /* Fill in the frame info */
      memset( (void *)fri, 0, sizeof( struct FrameInfo ) );

      fri -> fri_PropertyFlags       = di . PropertyFlags;
      fri -> fri_Resolution          = *(&(di . Resolution));
      fri -> fri_RedBits             = di . RedBits;
      fri -> fri_GreenBits           = di . GreenBits;
      fri -> fri_BlueBits            = di . BlueBits;
      fri -> fri_Dimensions . Width  = scr -> Width;
      fri -> fri_Dimensions . Height = scr -> Height;
      fri -> fri_Dimensions . Depth  = scr -> BitMap . Depth;
      fri -> fri_Screen              = scr;
      fri -> fri_ColorMap            = scr -> ViewPort . ColorMap;

      /* Send the message */
      dtf . MethodID          = DTM_FRAMEBOX;
      dtf . dtf_GInfo         = NULL;
      dtf . dtf_ContentsInfo  = fri;
      dtf . dtf_FrameInfo     = NULL;
      dtf . dtf_SizeFrameInfo = sizeof(struct FrameInfo);
      dtf . dtf_FrameFlags    = FRAMEF_SPECIFY;

      return( DoDTMethodA( dto, win, req, (Msg)(&dtf) ) );
    }

    return( 0UL );
}


static
void PrintFrameInfo( struct FrameInfo *fri )
{
    Printf( "fri: PropertyFlags : 0x%lx\n", (ULONG)(fri -> fri_PropertyFlags) );
    Printf( "fri: RedBits       : 0x%lx\n", (ULONG)(fri -> fri_RedBits) );
    Printf( "fri: GreenBits     : 0x%lx\n", (ULONG)(fri -> fri_GreenBits) );
    Printf( "fri: BlueBits      : 0x%lx\n", (ULONG)(fri -> fri_BlueBits) );
    Printf( "fri: Width         : %lu\n",   (ULONG)(fri -> fri_Dimensions . Width) );
    Printf( "fri: Height        : %lu\n",   (ULONG)(fri -> fri_Dimensions . Height) );
    Printf( "fri: Depth         : %lu\n",   (ULONG)(fri -> fri_Dimensions . Depth) );
    Printf( "fri: Screen        : 0x%lx\n", (ULONG)(fri -> fri_Screen) );
    Printf( "fri: ColorMap      : 0x%lx\n", (ULONG)(fri -> fri_ColorMap) );

    if( (fri -> fri_PropertyFlags) & DIPF_IS_HAM )
    {
      Printf( "fri: HAM\n" );
    }

    if( (fri -> fri_PropertyFlags) & DIPF_IS_EXTRAHALFBRITE )
    {
      Printf( "fri: ExtraHalfBrite\n" );
    }

    if( (fri -> fri_Flags) & FIF_SCALABLE )
    {
      Printf( "fri: FIF_SCALABLE, object can scale contents\n" );
    }

    if( (fri -> fri_Flags) & FIF_SCROLLABLE )
    {
      Printf( "fri: FIF_SCROLLABLE, object can scroll contents\n" );
    }

    if( (fri -> fri_Flags) & FIF_REMAPPABLE )
    {
      Printf( "fri: FIF_REMAPPABLE, object can remap contents\n" );
    }
}


static
BOOL FillTriggerData( struct dtTrigger *dtt, UBYTE *buffer, ULONG buffersize )
{
    ULONG type    = GetDTTriggerMethodDataFlags( (dtt -> dtt_Function) );
    BOOL  success = FALSE;

    if( dtt && buffer && buffersize )
    {
      dtt -> dtt_Data = NULL;

      if( type == STMD_VOID )
      {
        success = TRUE;
      }
      else
      {
        Printf( "enter trigger method data " );

        switch( type )
        {
          case STMD_ULONG:   Printf( "(ULONG)"    ); break;
          case STMD_STRPTR:  Printf( "(STRPTR)"   ); break;
          case STMD_TAGLIST: Printf( "(TAGLIST)"  ); break;
          default:           Printf( "--UNKOWN--" ); break;
        }

        FlushInOut();

        memset( buffer, 0, (size_t)buffersize );

        if( ReadItem( buffer, (buffersize - 1L), NULL ) )
        {
          switch( type )
          {
            case STMD_ULONG:
            {
                long longbuf;

                if( StrToLong( buffer, (&longbuf) ) != (-1L) )
                {
                  dtt -> dtt_Data = (APTR)longbuf;

                  success = TRUE;

                  Printf( "data %ld\n", longbuf );
                }
            }
                break;

            case STMD_STRPTR:
            {
                dtt -> dtt_Data = (APTR)buffer;

                Printf( "data \"%s\"\n", buffer );

                success = TRUE;
            }
                break;

            default: /* STMD_TAGLIST and others */
            {
                Printf( "trigger data not supported\n" );
            }
                break;
          }
        }

        /* flush input (see ReadItem autodoc) */
        FlushInOut();
      }
    }

    return( success );
}


static
void FlushInOut( void )
{
    Flush( Input() );
    Flush( Output() );
}



